#include <stdbool.h>
#include <stdint.h>
#include <array>
#include "BME280.h"
#include "MAX17043.h"

#ifdef __cplusplus
extern "C" {
#endif

#include "zboss_api.h"
#include "zb_mem_config_min.h"
#include "zb_zcl_reporting.h"
#include "zb_zcl_temp_measurement.h"
#include "zb_zcl_rel_humidity_measurement.h"
#include "zb_zcl_pressure_measurement.h"
#include "zb_zcl_identify.h"
#include "zb_zcl_basic.h"
#include "zb_error_handler.h"
#include "zb_nrf52840_internal.h"
#include "zb_multi_sensor.h"
#include "nrf_802154.h"

#include "app_timer.h"
#include "nrf_delay.h"
#include "nrf_gpio.h"
#include "boards.h"
#include "bsp.h"
#include "nrfx_twim.h"
    
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"
	
#ifdef __cplusplus
}
#endif

#define IEEE_CHANNEL_MASK                  (1l << ZIGBEE_CHANNEL)               /**< Scan only one, predefined channel to find the coordinator. */
#define ERASE_PERSISTENT_CONFIG            ZB_FALSE 
constexpr uint32_t ATT_REFRESH_RATE = 60000;


//constexpr zb_uint32_t IEEE_CHANNEL_MASK = ((1l << 25) | (1l << 24) | (1l << 20) | (1l << 19) | (1l << 15) | (1l << 14) | (1l << 11));
//constexpr zb_uint32_t IEEE_CHANNEL_MASK = (1l << 11);

static sensor_device_ctx_t m_dev_ctx { 0 };

ZB_ZCL_DECLARE_IDENTIFY_ATTRIB_LIST(identify_attr_list, &m_dev_ctx.identify_attr.identify_time);

ZB_ZCL_DECLARE_BASIC_ATTRIB_LIST_EXT(basic_attr_list,
    &m_dev_ctx.basic_attr.zcl_version,
    &m_dev_ctx.basic_attr.app_version,
    &m_dev_ctx.basic_attr.stack_version,
    &m_dev_ctx.basic_attr.hw_version,
    m_dev_ctx.basic_attr.mf_name,
    m_dev_ctx.basic_attr.model_id,
    m_dev_ctx.basic_attr.date_code,
    &m_dev_ctx.basic_attr.power_source,
    m_dev_ctx.basic_attr.location_id,
    &m_dev_ctx.basic_attr.ph_env,
    &m_dev_ctx.basic_attr.sw_ver);


ZB_ZCL_DECLARE_TEMP_MEASUREMENT_ATTRIB_LIST(temperature_attr_list, 
    &m_dev_ctx.temp_attr.measure_value,
    &m_dev_ctx.temp_attr.min_measure_value, 
    &m_dev_ctx.temp_attr.max_measure_value, 
    &m_dev_ctx.temp_attr.tolerance);

ZB_ZCL_DECLARE_PRES_MEASUREMENT_ATTRIB_LIST(pressure_attr_list, 
    &m_dev_ctx.pres_attr.measure_value, 
    &m_dev_ctx.pres_attr.min_measure_value, 
    &m_dev_ctx.pres_attr.max_measure_value, 
    &m_dev_ctx.pres_attr.tolerance);

ZB_ZCL_DECLARE_REL_HUMIDITY_MEASUREMENT_ATTRIB_LIST(humid_attr_list, 
    &m_dev_ctx.humid_attr.measure_value, 
    &m_dev_ctx.humid_attr.min_measure_value, 
    &m_dev_ctx.humid_attr.max_measure_value);

ZB_ZCL_DECLARE_POWER_CONFIG_BATTERY_ATTRIB_LIST_EXT(power_attr_list,
    &m_dev_ctx.power_attr.voltage_value,
    nullptr,                            
    nullptr,                            
    nullptr,                                        
    nullptr,
    nullptr,
    &m_dev_ctx.power_attr.remaining_value,
    nullptr,
    nullptr,
    nullptr,
    nullptr,
    nullptr,
    nullptr,
    nullptr,
    nullptr);

ZB_DECLARE_MULTI_SENSOR_CLUSTER_LIST(multi_sensor_clusters,
    basic_attr_list,
    identify_attr_list,
    temperature_attr_list,
    pressure_attr_list,
    humid_attr_list,
    power_attr_list);

ZB_ZCL_DECLARE_MULTI_SENSOR_EP(multi_sensor_ep,
    MULTI_SENSOR_ENDPOINT,
    multi_sensor_clusters);

ZBOSS_DECLARE_DEVICE_CTX_1_EP(multi_sensor_ctx, multi_sensor_ep);

/**@brief Function for initializing all clusters attributes.
 */
static void multi_sensor_clusters_attr_init(void)
{
    /* Basic cluster attributes data */
    m_dev_ctx.basic_attr.zcl_version   = ZB_ZCL_VERSION;
    m_dev_ctx.basic_attr.app_version   = SENSOR_INIT_BASIC_APP_VERSION;
    m_dev_ctx.basic_attr.stack_version = SENSOR_INIT_BASIC_STACK_VERSION;
    m_dev_ctx.basic_attr.hw_version    = SENSOR_INIT_BASIC_HW_VERSION;

    /* Use ZB_ZCL_SET_STRING_VAL to set strings, because the first byte should
     * contain string length without trailing zero.
     *
     * For example "test" string wil be encoded as:
     *   [(0x4), 't', 'e', 's', 't']
     */
    ZB_ZCL_SET_STRING_VAL(m_dev_ctx.basic_attr.mf_name,
        SENSOR_INIT_BASIC_MANUF_NAME,
        ZB_ZCL_STRING_CONST_SIZE(SENSOR_INIT_BASIC_MANUF_NAME));

    ZB_ZCL_SET_STRING_VAL(m_dev_ctx.basic_attr.model_id,
        SENSOR_INIT_BASIC_MODEL_ID,
        ZB_ZCL_STRING_CONST_SIZE(SENSOR_INIT_BASIC_MODEL_ID));

    ZB_ZCL_SET_STRING_VAL(m_dev_ctx.basic_attr.date_code,
        SENSOR_INIT_BASIC_DATE_CODE,
        ZB_ZCL_STRING_CONST_SIZE(SENSOR_INIT_BASIC_DATE_CODE));

    m_dev_ctx.basic_attr.power_source = SENSOR_INIT_BASIC_POWER_SOURCE;

    ZB_ZCL_SET_STRING_VAL(m_dev_ctx.basic_attr.location_id,
        SENSOR_INIT_BASIC_LOCATION_DESC,
        ZB_ZCL_STRING_CONST_SIZE(SENSOR_INIT_BASIC_LOCATION_DESC));


    m_dev_ctx.basic_attr.ph_env = SENSOR_INIT_BASIC_PH_ENV;

    /* Identify cluster attributes data */
    m_dev_ctx.identify_attr.identify_time        = ZB_ZCL_IDENTIFY_IDENTIFY_TIME_DEFAULT_VALUE;

    /* Temperature measurement cluster attributes data */
    m_dev_ctx.temp_attr.measure_value            = ZB_ZCL_ATTR_TEMP_MEASUREMENT_VALUE_UNKNOWN;
    m_dev_ctx.temp_attr.min_measure_value        = ZB_ZCL_ATTR_TEMP_MEASUREMENT_MIN_VALUE_MIN_VALUE;
    m_dev_ctx.temp_attr.max_measure_value        = ZB_ZCL_ATTR_TEMP_MEASUREMENT_MAX_VALUE_MAX_VALUE;
    m_dev_ctx.temp_attr.tolerance                = ZB_ZCL_ATTR_TEMP_MEASUREMENT_TOLERANCE_MAX_VALUE;

    /* Pressure measurement cluster attributes data */
    m_dev_ctx.pres_attr.measure_value            = ZB_ZCL_ATTR_PRES_MEASUREMENT_VALUE_UNKNOWN;
    m_dev_ctx.pres_attr.min_measure_value        = ZB_ZCL_ATTR_PRES_MEASUREMENT_MIN_VALUE_MIN_VALUE;
    m_dev_ctx.pres_attr.max_measure_value        = ZB_ZCL_ATTR_PRES_MEASUREMENT_MAX_VALUE_MAX_VALUE;
    m_dev_ctx.pres_attr.tolerance                = ZB_ZCL_ATTR_PRES_MEASUREMENT_TOLERANCE_MAX_VALUE;
    
    /* Humidity measurement cluster attributes data */
    m_dev_ctx.humid_attr.measure_value            = ZB_ZCL_ATTR_REL_HUMIDITY_MEASUREMENT_VALUE_UNKNOWN;
    m_dev_ctx.humid_attr.min_measure_value        = ZB_ZCL_ATTR_REL_HUMIDITY_MEASUREMENT_MIN_VALUE_MIN_VALUE;
    m_dev_ctx.humid_attr.max_measure_value        = ZB_ZCL_ATTR_REL_HUMIDITY_MEASUREMENT_MAX_VALUE_MAX_VALUE;
    m_dev_ctx.humid_attr.tolerance				  = ZB_ZCL_ATTR_REL_HUMIDITY_MEASUREMENT_VALUE_UNKNOWN;
    
    /* Power Config cluster attribute data */
    m_dev_ctx.power_attr.voltage_value = ZB_ZCL_POWER_CONFIG_BATTERY_VOLTAGE_INVALID;
    m_dev_ctx.power_attr.remaining_value = ZB_ZCL_POWER_CONFIG_BATTERY_REMAINING_UNKNOWN;
}

//#ifdef __cplusplus
//}
//#endif

BME280 bme;
void initBME280()
{
    nrfx_twim_t twim = { .p_twim = NRF_TWIM0, .drv_inst_idx = NRFX_TWIM0_INST_IDX };
    const nrfx_twim_config_t twim_config = { 
        .scl = ARDUINO_SCL_PIN,
        .sda = ARDUINO_SDA_PIN,
        .frequency = NRF_TWIM_FREQ_400K,
        .interrupt_priority = APP_IRQ_PRIORITY_MID,
        .hold_bus_uninit = false
    };
    
    bme.init(twim, twim_config, BME280_I2C_ADDR_PRIM, false);
    bme.setSensorSettings(MODE_FORCED, BME280_OVERSAMPLING_1X, BME280_OVERSAMPLING_1X, BME280_OVERSAMPLING_1X, BME280_FILTER_COEFF_OFF, BME280_STANDBY_TIME_1_MS);
}

MAX17043 fuel_gauge;
void initFuelGauge()
{
    nrfx_twim_t twim_fg = { .p_twim = NRF_TWIM1, .drv_inst_idx = NRFX_TWIM1_INST_IDX };
    const nrfx_twim_config_t twim_config_fg = { 
        .scl = ARDUINO_3_PIN,
        .sda = ARDUINO_2_PIN,
        .frequency = NRF_TWIM_FREQ_400K,
        .interrupt_priority = APP_IRQ_PRIORITY_MID,
        .hold_bus_uninit = false
    };
    
    fuel_gauge.init(twim_fg, twim_config_fg, MAX17043_I2C_ADDR);
    fuel_gauge.powerOnReset();
    fuel_gauge.quickStart();
    nrf_delay_ms(1000);
}

static app_timer_t zb_app_timer_data = { { 0 } };
static const app_timer_id_t zb_app_timer = &zb_app_timer_data;

void init_timer()
{
    ret_code_t err_code = app_timer_init();
    APP_ERROR_CHECK(err_code);
}

void init_log()
{
    ret_code_t err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);
    
    NRF_LOG_DEFAULT_BACKENDS_INIT();
}

/**@brief Function for initializing LEDs.
 */
void gpio_init(void)
{
    /* Initialize LEDs and buttons - use BSP to control them. */
    //bsp_board_init(BSP_INIT_LEDS);
    nrf_gpio_cfg_output(LED_1);
    nrf_gpio_pin_write(LED_1, 0);
    
    nrf_gpio_cfg_sense_input(11, NRF_GPIO_PIN_PULLDOWN, NRF_GPIO_PIN_SENSE_LOW);
    
    // Configure antenna switch GPIO P0.24 & P0.25 
    nrf_gpio_range_cfg_output(24, 25);
}

/**@brief Function for configuring the onboard antenna switch
 * 
 * @param[IN]   useExternalAnt   True = use external antenna, False = use onboard chip antenna.
 * 
 * @details 
 */
void select_antenna(bool useExternalAnt)
{
    if (useExternalAnt)
    {
        nrf_gpio_pin_write(24, 0);
        nrf_gpio_pin_write(25, 1);
        
    }
    else
    {
        nrf_gpio_pin_write(24, 1);
        nrf_gpio_pin_write(25, 0);
    }
}

void attribute_update()
{
    zb_zcl_status_t zcl_status;
    static zb_int16_t new_temp_value, new_pres_value, new_humid_value { 0 };
    static zb_uint8_t new_batt_pct { 0 };
    
    // Get new sensor readings
    bme280_data data = { 0 };
    bme.getSensorData(BME280_ALL, data, true);

    /* Get new temperature measured value */
    new_temp_value = static_cast<zb_int16_t>(data.temperature);
    zcl_status = zb_zcl_set_attr_val(MULTI_SENSOR_ENDPOINT, 
        ZB_ZCL_CLUSTER_ID_TEMP_MEASUREMENT, 
        ZB_ZCL_CLUSTER_SERVER_ROLE, 
        ZB_ZCL_ATTR_TEMP_MEASUREMENT_VALUE_ID, 
        (zb_uint8_t *)&new_temp_value, 
        ZB_FALSE);
    NRF_LOG_INFO("Set temperature value set to: %d", new_temp_value);
    if (zcl_status != ZB_ZCL_STATUS_SUCCESS)
    {
        NRF_LOG_INFO("Set temperature value fail. zcl_status: %d", zcl_status);
    }

    /* Get new pressure measured value */
//    new_pres_value = static_cast<zb_int16_t>(data.pressure / 25600);
//    zcl_status = zb_zcl_set_attr_val(MULTI_SENSOR_ENDPOINT,
//        ZB_ZCL_CLUSTER_ID_PRESSURE_MEASUREMENT, 
//        ZB_ZCL_CLUSTER_SERVER_ROLE, 
//        ZB_ZCL_ATTR_PRES_MEASUREMENT_VALUE_ID, 
//        (zb_uint8_t *)&new_pres_value, 
//        ZB_FALSE);
//    NRF_LOG_INFO("Set pressure value set to: %d", new_pres_value);
//    if (zcl_status != ZB_ZCL_STATUS_SUCCESS)
//    {
//        NRF_LOG_INFO("Set pressure value fail. zcl_status: %d", zcl_status);
//    }
    
    /* Get new humidity measured value */
    new_humid_value = static_cast<zb_uint16_t>(static_cast<float>((data.humidity / 1024.0)) * 100);
    zcl_status = zb_zcl_set_attr_val(MULTI_SENSOR_ENDPOINT,
        ZB_ZCL_CLUSTER_ID_REL_HUMIDITY_MEASUREMENT, 
        ZB_ZCL_CLUSTER_SERVER_ROLE, 
        ZB_ZCL_ATTR_REL_HUMIDITY_MEASUREMENT_VALUE_ID, 
        (zb_uint8_t *)&new_humid_value, 
        ZB_FALSE);
    NRF_LOG_INFO("Set humidity value set to: %d", new_humid_value);
    if (zcl_status != ZB_ZCL_STATUS_SUCCESS)
    {
        NRF_LOG_INFO("Set humidity value fail. zcl_status: %d", zcl_status);
    }
    
    /* Get new battery measured value */
    new_batt_pct = static_cast<uint8_t>(fuel_gauge.readPercentage());
    zcl_status = zb_zcl_set_attr_val(MULTI_SENSOR_ENDPOINT,
        ZB_ZCL_CLUSTER_ID_POWER_CONFIG, 
        ZB_ZCL_CLUSTER_SERVER_ROLE, 
        ZB_ZCL_ATTR_POWER_CONFIG_BATTERY_REMAINING_ID, 
        (zb_uint8_t *)&new_batt_pct, 
        ZB_FALSE);
    NRF_LOG_INFO("Battery remaining set to: %d", new_batt_pct);
    if (zcl_status != ZB_ZCL_STATUS_SUCCESS)
    {
        NRF_LOG_INFO("Set battery remaining fail. zcl_status: %d", zcl_status);
    }
    fuel_gauge.setSleepMode(true);
}

/**@brief Function for handling nrf app timer.
 * 
 * @param[IN]   context   Void pointer to context function is called with.
 * 
 * @details Function is called with pointer to sensor_device_ep_ctx_t as argument.
 */
//static void zb_app_timer_handler(void * context)
void zb_app_timer_handler(zb_uint8_t param)
{
    attribute_update();
    
    zb_ret_t zb_err_code = zb_schedule_alarm(zb_app_timer_handler, 1, 60 * ZB_TIME_ONE_SECOND);
    APP_ERROR_CHECK(zb_err_code);
}

#ifdef __cplusplus
extern "C" {
#endif

    /**
     * @brief callback function for handling ZCL commands.
     * 
     * @param[in]   param   Reference to Zigbee stack buffer used to pass received data.
     */
    static zb_void_t zcl_device_cb(zb_uint8_t param)
    {
    
    }

    /**
     * @brief Perform local operation - Leave network.
     * 
     * param[in]    param   Reference to ZigBee stack buffer that will be used to construct leave request.
     **/
    zb_void_t zigbee_leave_network(zb_uint8_t param)
    {
        zb_ret_t zb_err_code;
    
        // Leaving network
        if(param)
        {
            zb_buf_t *p_buf = ZB_BUF_FROM_REF(param);
            zb_zdo_mgmt_leave_param_t *p_req_param;
        
            p_req_param = ZB_GET_BUF_PARAM(p_buf, zb_zdo_mgmt_leave_param_t);
            UNUSED_RETURN_VALUE(ZB_BZERO(p_req_param, sizeof(zb_zdo_mgmt_leave_param_t)));
        
            // Set dst_addr == local address for local leave.
            p_req_param->dst_addr = ZB_PIBCACHE_NETWORK_ADDRESS();
            p_req_param->rejoin = ZB_FALSE;
            UNUSED_RETURN_VALUE(zdo_mgmt_leave_req(param, NULL));
        }
        else
        {
            zb_err_code = zb_get_out_buf_delayed(zigbee_leave_network);
            ZB_ERROR_CHECK(zb_err_code);
        }
    }

    /**
     * @brief Function for starting join/rejoin procedure.
     * 
     * param[in]    leave_type  Type of leave request (with or without rejoin).
     **/
    zb_void_t zigbee_retry_join(zb_uint8_t leave_type)
    {
        zb_bool_t comm_status;
    
        if (leave_type == ZB_NWK_LEAVE_TYPE_RESET)
        {
            comm_status = bdb_start_top_level_commissioning(ZB_BDB_NETWORK_STEERING);
            ZB_COMM_STATUS_CHECK(comm_status);
        }
    }

    zb_void_t zigbee_leave_and_join(zb_uint8_t param)
    {
        if (ZB_JOINED())
        {
            // Leave network. Joining procedure will be initiated inside ZigBee stack handler.
            zigbee_leave_network(param);
        }
        else
        {
            zigbee_retry_join(ZB_NWK_LEAVE_TYPE_RESET);
        
            if (param)
            {
                ZB_FREE_BUF_BY_REF(param);
            }
        }
    }
    
    void configure_attribute_reporting()
    {
        zb_zcl_reporting_info_t temp_rep_info;
        memset(&temp_rep_info, 0, sizeof(temp_rep_info));
        
        temp_rep_info.direction = ZB_ZCL_CONFIGURE_REPORTING_SEND_REPORT;
        temp_rep_info.ep = MULTI_SENSOR_ENDPOINT;
        temp_rep_info.cluster_id = ZB_ZCL_CLUSTER_ID_TEMP_MEASUREMENT;
        temp_rep_info.cluster_role = ZB_ZCL_CLUSTER_SERVER_ROLE;
        temp_rep_info.attr_id = ZB_ZCL_ATTR_TEMP_MEASUREMENT_VALUE_ID;
        temp_rep_info.dst.profile_id = ZB_AF_HA_PROFILE_ID;
        temp_rep_info.u.send_info.min_interval = 30;      // 30 seconds
        temp_rep_info.u.send_info.max_interval = 300;    // 5 minutes
        temp_rep_info.u.send_info.delta.s16 = 0x0032;        // 0.5 degrees
        
        zb_zcl_reporting_info_t humid_rep_info;
        memset(&humid_rep_info, 0, sizeof(humid_rep_info));
        
        humid_rep_info.direction = ZB_ZCL_CONFIGURE_REPORTING_SEND_REPORT;
        humid_rep_info.ep = MULTI_SENSOR_ENDPOINT;
        humid_rep_info.cluster_id = ZB_ZCL_CLUSTER_ID_REL_HUMIDITY_MEASUREMENT;
        humid_rep_info.cluster_role = ZB_ZCL_CLUSTER_SERVER_ROLE;
        humid_rep_info.attr_id = ZB_ZCL_ATTR_REL_HUMIDITY_MEASUREMENT_VALUE_ID;
        humid_rep_info.dst.profile_id = ZB_AF_HA_PROFILE_ID;
        humid_rep_info.u.send_info.min_interval = 60;       // 60 seconds
        humid_rep_info.u.send_info.max_interval = 600;     // 10 minutes
        humid_rep_info.u.send_info.delta.u16 = 0x0064;         // 1% rh
        
        zb_zcl_reporting_info_t batt_rep_info;
        memset(&batt_rep_info, 0, sizeof(batt_rep_info));
        
        batt_rep_info.direction = ZB_ZCL_CONFIGURE_REPORTING_SEND_REPORT;
        batt_rep_info.ep = MULTI_SENSOR_ENDPOINT;
        batt_rep_info.cluster_id = ZB_ZCL_CLUSTER_ID_POWER_CONFIG;
        batt_rep_info.cluster_role = ZB_ZCL_CLUSTER_SERVER_ROLE;
        batt_rep_info.attr_id = ZB_ZCL_ATTR_POWER_CONFIG_BATTERY_REMAINING_ID;
        batt_rep_info.dst.profile_id = ZB_AF_HA_PROFILE_ID;
        batt_rep_info.u.send_info.min_interval = 30;       // 30 seconds
        batt_rep_info.u.send_info.max_interval = 21600;     // 5 minutes
        batt_rep_info.u.send_info.delta.u8 = 0x01;         // 1%
        
        zb_zcl_put_reporting_info(&temp_rep_info, ZB_TRUE);
        zb_zcl_put_reporting_info(&humid_rep_info, ZB_TRUE);
        zb_zcl_put_reporting_info(&batt_rep_info, ZB_TRUE);       
    }

    /**
     * @brief ZigBee stack event handler.
     * 
     * param[in]    param   Reference to ZigBee stack buffer used to pass arguments (signal).27.46
     **/
    zb_void_t zboss_signal_handler(zb_uint8_t param)
    {
        zb_zdo_app_signal_hdr_t *p_sg_p = NULL;
        zb_zdo_signal_leave_params_t *p_leave_params = NULL;
        zb_zdo_app_signal_type_t sig = zb_get_app_signal(param, &p_sg_p);
        zb_ret_t status = ZB_GET_APP_SIGNAL_STATUS(param);
        zb_ret_t zb_err_code;
    
        switch (sig)
        {
        case ZB_BDB_SIGNAL_DEVICE_FIRST_START:
        case ZB_BDB_SIGNAL_DEVICE_REBOOT:
            {
                if (status == RET_OK)
                {
                    NRF_LOG_INFO("Joined network successfully");
                    nrf_gpio_pin_write(LED_1, 0);
                    attribute_update();
//                    ret_code_t err_code = app_timer_start(zb_app_timer, APP_TIMER_TICKS(ATT_REFRESH_RATE), NULL);
                    zb_err_code = zb_schedule_alarm(zb_app_timer_handler, 1, 60 * ZB_TIME_ONE_SECOND);
                    APP_ERROR_CHECK(zb_err_code);
                    configure_attribute_reporting();
                    zb_zdo_pim_set_long_poll_interval(15000);
                }
                else
                {
                    NRF_LOG_ERROR("Failed to join network. Status: %d", status);
                    nrf_gpio_pin_write(LED_1, 1);
                    zb_err_code = bdb_start_top_level_commissioning(ZB_BDB_NETWORK_STEERING);
                    ZB_COMM_STATUS_CHECK(zb_err_code);
                    //zb_err_code = ZB_SCHEDULE_ALARM(zigbee_leave_and_join, 0, ZB_TIME_ONE_SECOND);
                    //ZB_ERROR_CHECK(zb_err_code);
                }
            }
            break;
        
        case ZB_ZDO_SIGNAL_LEAVE:
            {
                if (status == RET_OK)
                {
                    nrf_gpio_pin_write(LED_1, 1);
                    p_leave_params = ZB_ZDO_SIGNAL_GET_PARAMS(p_sg_p, zb_zdo_signal_leave_params_t);
                    NRF_LOG_INFO("Network left. Leave type: %d", p_leave_params->leave_type);
                    // Retry join. 
                    zigbee_retry_join(p_leave_params->leave_type);   // Should this equal ZB_NWK_LEAVE_TYPE_RESET? Thus call zigbee_leave_network
                
//                    ret_code_t err_code = app_timer_stop(zb_app_timer);
//                    APP_ERROR_CHECK(err_code);
                    
                    zb_err_code = zb_schedule_alarm_cancel(zb_app_timer_handler, 1, nullptr);
                    ZB_ERROR_CHECK(zb_err_code);
                }
                else
                {
                    NRF_LOG_ERROR("Unable to leave network. Status: %d", status);
                }
            }
            break;
        
        case ZB_ZDO_SIGNAL_DEVICE_ANNCE:
            {
                zb_zdo_signal_device_annce_params_t * dev_annce_params = ZB_ZDO_SIGNAL_GET_PARAMS(p_sg_p, zb_zdo_signal_device_annce_params_t);
                NRF_LOG_INFO("Device with a short address %hx commissionned", dev_annce_params->device_short_addr);
            }
            break;
        
        case ZB_COMMON_SIGNAL_CAN_SLEEP:
            {
                zb_zdo_signal_can_sleep_params_t *can_sleep_params = ZB_ZDO_SIGNAL_GET_PARAMS(p_sg_p, zb_zdo_signal_can_sleep_params_t);
                NRF_LOG_INFO("Can sleep for %ld ms", can_sleep_params->sleep_tmo);
                zb_sleep_now();
            }
            break;
        
        case ZB_ZDO_SIGNAL_PRODUCTION_CONFIG_READY:
            {
                if (status != RET_OK)
                {
                    NRF_LOG_WARNING("Production config is not present or invalid.");
                }
            }
            break;
        
        default:
            // Unhandled signal. For more information see: zb_zdo_app_signal_type_e and zb_ret_e
            NRF_LOG_INFO("Unhandled signal %d. Status: %d", sig, status);
        }
    
        if (param)
        {
            ZB_FREE_BUF_BY_REF(param);
        }
    }
    
    zb_void_t sleepy_device_setup()
    {
        if (nrf_gpio_pin_read(11) == 0)
        {
            NRF_LOG_INFO("RX on when idle is false");
            zb_set_rx_on_when_idle(ZB_FALSE);
        }
        else
        {
            NRF_LOG_INFO("RX on when idle is true");
            zb_set_rx_on_when_idle(ZB_TRUE);
        }
    }
    
    //    zb_void_t report_attribute_cb(zb_zcl_addr_t *addr, zb_uint8_t ep, zb_uint16_t cluster_id, zb_uint16_t attr_id, zb_uint8_t attr_type, zb_uint8_t *value)
    //    {
    //        NRF_LOG_INFO("Attribute reported. Cluster ID: %d \t Attribute ID: %d \t Value: %d", cluster_id, attr_id, value);
    //    }
    
    
    zb_void_t zigbee_init()
    {
        zb_ret_t zb_err_code;
        ret_code_t err_code;
        zb_ieee_addr_t ieee_addr;
        
        // Set Zigbee Stack logging level and traffic dump subsystem
        ZB_SET_TRACE_LEVEL(ZIGBEE_TRACE_LEVEL);
        ZB_SET_TRACE_MASK(ZIGBEE_TRACE_MASK);
        ZB_SET_TRAF_DUMP_OFF();
    
        // Initialize Zigbee stack
        ZB_INIT("multi_sensor");
    
        // Set device address to the value read from FICR registers
        zb_osif_get_ieee_eui64(ieee_addr);
        zb_set_long_address(ieee_addr);
    
        // Set static long IEEE address.
        zb_set_network_ed_role(IEEE_CHANNEL_MASK);
        zb_set_nvram_erase_at_start(ERASE_PERSISTENT_CONFIG);
        //zigbee_erase_persistent_storage(ERASE_PERSISTENT_CONFIG);
    
        zb_set_ed_timeout(ED_AGING_TIMEOUT_64MIN);
        zb_set_keepalive_timeout(ZB_MILLISECONDS_TO_BEACON_INTERVAL(300000));
        zb_set_rx_on_when_idle(ZB_FALSE);
        //sleepy_device_setup();
    
        // Initialize application context structure
        //UNUSED_RETURN_VALUE(ZB_MEMSET(&m_dev_ctx, 0, sizeof(m_dev_ctx))); // Initialized when declared. No need to do a memset.
    
        // Register sensor device context (endpoints)
        //ZB_AF_REGISTER_DEVICE_CTX(&multi_sensor_ctx);
        zb_af_register_device_ctx(&multi_sensor_ctx);
    
        // Initialize sensor device attributes
        multi_sensor_clusters_attr_init();
        
        //ZB_ZCL_SET_REPORT_ATTR_CB(&report_attribute_cb);
    
        // Start Zigbee stack
        zb_err_code = zboss_start();
        ZB_ERROR_CHECK(zb_err_code);
    }
#ifdef __cplusplus
}
#endif

int main(void)
{
    NRF_POWER->DCDCEN = 1;
    
    init_log();
//    init_timer();
    gpio_init();
    select_antenna(true);
    initBME280();
    initFuelGauge();
    
    // Create timer for reporting attribute;
    // = app_timer_create(&zb_app_timer, APP_TIMER_MODE_REPEATED, zb_app_timer_handler);
    //APP_ERROR_CHECK(err_code);
    
    // Set TX Power
    nrf_802154_tx_power_set(0);
    //NRF_LOG_INFO("Current TX power level: %d", nrf_802154_tx_power_get());
    
    zigbee_init();
    
//    configure_attribute_reporting();
    
//    NRF_LOG_INFO("Temperature Cluster Reporting? %d", zcl_is_attr_reported(10, ZB_ZCL_CLUSTER_ID_TEMP_MEASUREMENT, ZB_ZCL_CLUSTER_SERVER_ROLE, 0x0));
//    NRF_LOG_INFO("Pressure Cluster Reporting? %d", zcl_is_attr_reported(10, ZB_ZCL_CLUSTER_ID_PRESSURE_MEASUREMENT, ZB_ZCL_CLUSTER_SERVER_ROLE, 0x0));
//    zb_zcl_start_attr_reporting(10, ZB_ZCL_CLUSTER_ID_TEMP_MEASUREMENT, ZB_ZCL_CLUSTER_SERVER_ROLE, ZB_ZCL_ATTR_TEMP_MEASUREMENT_VALUE_ID);
  
    
    //Max17043

//    uint16_t volts = fuel_gauge.readVoltage();
//    uint8_t soc = static_cast<uint8_t >(fuel_gauge.readPercentage());
//    fuel_gauge.setSleepMode(true);
//    
//    NRF_LOG_INFO("Battery SOC: %d", soc);
//    NRF_LOG_PROCESS();
//    nrf_delay_ms(2000);
    
    while (1)
    {
        zboss_main_loop_iteration();
        UNUSED_RETURN_VALUE(NRF_LOG_PROCESS());
    }
}
